//***********************************************************//
//  Pluck.cpp: Implementation of the Pluck Object            //
//                                                           //
//                                                           //
//                                                           //
//***********************************************************//

#include "Pluck.h"

//////////CONSTRUCTION /////////////////////

Pluck::Pluck(){ 

m_inputfr = 0;
m_amp = 1.f;
m_fr= 440.f;
m_maxscale = 32767.f;

float tdelay = m_sr/m_fr;
int delay = (int) (tdelay -.5);
float fracdelay = (float)(tdelay - (delay +.5));
m_a = (1-fracdelay)/(1+fracdelay);

m_vdtime = m_size - tdelay;

m_input = new Randh((m_amp*10000.f/m_maxscale), m_amp, 0, 0, 1, m_sr);
m_fdbgain = 0.9f;

AddMsg("amplitude", 41);
AddMsg("maxscale", 42);

} 

Pluck::Pluck(float fr, float amp, float fdbgain, 
              SndObj* InFrObj, float maxscale, int vecsize, float sr)
	  :StringFlt(fr, fdbgain, 0, InFrObj, vecsize, sr)
{

m_amp = amp;
m_maxscale = maxscale;
m_input = new Randh((m_amp*10000.f/m_maxscale), m_amp, 0, 0, 1, m_sr);
FillDelay();
AddMsg("amplitude", 41);
AddMsg("maxscale", 42);;

}

Pluck::~Pluck()
{
delete m_input;
}

//////////////////OPERATIONS //////////////////////
void
Pluck::FillDelay(){
  for(int n=0; n< m_size;n++){
   m_input->DoProcess();
   PutSample(m_input->Output(0));
  }
  m_wpointer = 0;
}

int
Pluck::Set(char* mess, float value){

	switch (FindMsg(mess)){

	case 41:
    SetAmp(value, m_maxscale);
	return 1;

	case 42:
	SetAmp(m_amp, value);
    return 1;

	default:
    return StringFlt::Set(mess,value);
     
	}


}
void 
Pluck::RePluck(){
Disable();
FillDelay();
Enable();
}

short
Pluck::DoProcess(){

if(!m_error){

  float fr, pos, tdelay, fracdelay, w, s, output;
  int delay;
for(m_vecpos=0; m_vecpos<m_vecsize;m_vecpos++){
 if(m_enable){
       if(m_inputfr){
       fr = m_fr + m_inputfr->Output(m_vecpos);
       if(fr < 20.f) fr = 20.f;
       tdelay = m_sr/fr;
       delay = (int)(tdelay - .5);
       fracdelay = (tdelay  - (delay +.5f));
       m_vdtime = m_size - tdelay;
       m_a = (1-fracdelay)/(1+fracdelay);
          }
   
   pos = m_wpointer + m_vdtime;
   while(pos > m_size) pos -= m_size;
   while(pos < 0) pos += m_size;

   w = GetSample(pos); 
   s = m_LPDelay*.5f + w*.5f;
   m_LPDelay = w;
   output = m_APDelay + s*(m_a);
   m_APDelay = s + output*(-m_a);
   PutSample(output*m_fdbgain); 
   m_output[m_vecpos] = output;  
    } 
 else m_output[m_vecpos] = 0.f;
}
 return 1;
}
else return 0;

}










